/*
 * Copyright (c) 2001-2005 Atmark Techno, Inc.  All Rights Reserved.
 *
 * Commands to program the Fujitsu MBM29DL324BD flash chips on the
 * Armadillo board.  These are 32 Mbit "bottom boot"
 * (boot and param blocks at low end of address space) flash chips.
 */

#include <stddef.h>
#include <target/buffer.h>
#include <target/io.h>
#include <target/herrno.h>
#include <target/scan.h>
#include <byteorder/big_endian.h>
#include "flash.h"
#include "memregions.h"

#define FLASH_TIMEOUT 400000000

/* download buffer for the flash programmer */
#define DLBUFSIZE (8*1024)
unsigned char dlbuf [DLBUFSIZE];
word_t dlbufsize = DLBUFSIZE;


/* these two defines should be move to architecture specific header
 * once we generalized this flash code.     yashi Feb 06, 2004  */
#define read16(addr)                                                    \
        ({ unsigned short __v = (*(volatile unsigned short *) (addr)); __v; })
#define write16(addr, b)                                        \
        (void)((*(volatile unsigned short *) (addr)) = (b))

#define flash_read(addr)         read16(addr)
#define flash_write(addr, b)     write16((addr), (b))
#define flash_read_cmd(base, offset) cpu_to_le16(read16(base + (offset<<1)))
#define flash_write_cmd(addr, b) write16((addr), cpu_to_le16(b))

/* handy flash functions */
static void flash_print_chip_error_string(unsigned char status);
static void flash_print_error_string(void);
static unsigned short flash_status_wait(addr_t addr, unsigned short value);
static int flash_status_full_check(addr_t addr, unsigned short value1, unsigned short value2);

static void flash_print_chip_error_string(unsigned char status)
{
	hprint("TimeOut");
}

/*
 * Come up with an error string representing the flash error; we can
 * have distinct errors for the two chips.  Get the status from
 * errdata, since this is used for error output.
 */
static void flash_print_error_string(void)
{
	hprint("lowbank:");
	flash_print_chip_error_string(errdata.word);
	hprint(" highbank:");
	flash_print_chip_error_string(errdata.word >> 16);
}

/*
 * Loop until both write state machines complete.
 */
static unsigned short flash_status_wait(addr_t addr, unsigned short value)
{
	unsigned short status;
	long timeout = FLASH_TIMEOUT;

	while (((status = (flash_read(addr))) != value) && timeout > 0) {
                timeout--;
	}
	return status;
}

/*
 * Loop until the Write State machine is ready, then do a full error
 * check.  Clear status and leave the flash in Read Array mode; return
 * 0 for no error, -1 for error.
 */
static int flash_status_full_check(addr_t addr, unsigned short value1, unsigned short value2)
{
	unsigned short status1, status2;

	status1 = flash_status_wait(addr, value1);
	status2 = flash_status_wait(addr + 2, value2);
	if (status1 != value1 || status2 != value2) {
		errdata.word = status1 | status2 << 16;
		errfunc = &flash_print_error_string;
	}
	return (status1 != value1 || status2 != value2) ? -1 : 0;
}

/*
 * Program the contents of the download buffer to flash at the given
 * address.  Size is also specified; we shouldn't have to track usage
 * of the download buffer, since the host side can easily do that.
 *
 * We write words without checking status between each; we only go
 * through the full status check procedure once, when the full buffer
 * has been written.
 *
 * Alignment problems are errors here; we don't automatically correct
 * them because in the context of this command they signify bugs, and
 * we want to be extra careful when writing flash.
 */
static int flash_program_cmdfunc(int argc, char *argv[])
{
	addr_t addr;
	size_t size;
	unsigned short *ptr;
	int status = 0;

	if (argc != 3)
		return -H_EUSAGE;
	if (scan(*++argv, &addr)) return -H_EADDR;
	if (scan(*++argv, &size)) return -H_EADDR;
	if (addr & UNALIGNED_MASK) return -H_EALIGN;
	if (size & UNALIGNED_MASK) return -H_EALIGN;

	size >>= BYTE_TO_WORD_SHIFT;
	for (ptr = (unsigned short*)dlbuf; size--; addr += sizeof(word_t)) {
		flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0xAA);
		flash_write_cmd(FLASH_BASEADDR + 0x554, 0x55);
		flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0xA0);
		flash_write(addr, *ptr);
		flash_status_wait(addr, *ptr);
		flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0xAA);
		flash_write_cmd(FLASH_BASEADDR + 0x554, 0x55);
		flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0xA0);
		flash_write(addr + 2, *(ptr + 1));
		flash_status_wait(addr + 2, *(ptr + 1));
		status |= flash_status_full_check(addr, *ptr, *(ptr + 1));
		ptr += 2;
	}
	return status;
}

const command_t flash_program_command =
{ "program", "<addr> <size>",
  "program flash from download buffer",
  &flash_program_cmdfunc };

/*
 * Erase a flash block.  Single argument is the first address in the
 * block (actually, it can be anywhere in the block, but the host-side
 * hermit downloader gives the first address).
 */
static int flash_erase_cmdfunc(int argc, char *argv[])
{
	addr_t addr;

	if (argc != 2)
		return -H_EUSAGE;
	if (scan(*++argv, &addr)) return -H_EADDR;
	addr &= ~(addr_t)UNALIGNED_MASK;	/* word align */

	flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0xAA);
	flash_write_cmd(FLASH_BASEADDR + 0x554, 0x55);
	flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0x80);
	flash_write_cmd(FLASH_BASEADDR + 0xAAA, 0xAA);
	flash_write_cmd(FLASH_BASEADDR + 0x554, 0x55);
	flash_write_cmd(addr, 0x30);
	return flash_status_full_check(addr, 0xFFFF, 0xFFFF);
}

const command_t flash_erase_command =
{ "erase", "<addr>", "erase flash block", &flash_erase_cmdfunc };

/*
 * Get Flash Size
 */
int flash_get_size(void)
{
        int ret;

        flash_write_cmd(FLASH_BASEADDR, 0xf0);
        flash_write_cmd(FLASH_BASEADDR + 0xaa, 0x98);
        if (flash_read_cmd(FLASH_BASEADDR, 0x10) == 'Q' &&
            flash_read_cmd(FLASH_BASEADDR, 0x11) == 'R' &&
            flash_read_cmd(FLASH_BASEADDR, 0x12) == 'Y') {
                ret = flash_read_cmd(FLASH_BASEADDR, 0x27);
        } else {
                ret = -1;
        }
        flash_write_cmd(FLASH_BASEADDR, 0xf0);

        return ret;
}
